from functools import partial
from pathlib import Path
import random

from textual import on
from textual.app import ComposeResult
from textual import getters
from textual.binding import Binding
from textual.command import Hit, Hits, Provider, DiscoveryHit
from textual.content import Content
from textual.events import ScreenResume
from textual.screen import Screen
from textual.reactive import var, reactive
from textual.widgets import Footer, OptionList, DirectoryTree, Tree
from textual import containers
from textual.widget import Widget


from toad.app import ToadApp
from toad import messages
from toad.agent_schema import Agent
from toad.acp import messages as acp_messages

from toad.widgets.plan import Plan
from toad.widgets.throbber import Throbber
from toad.widgets.conversation import Conversation
from toad.widgets.project_directory_tree import ProjectDirectoryTree
from toad.widgets.side_bar import SideBar


class ModeProvider(Provider):
    async def search(self, query: str) -> Hits:
        """Search for Python files."""
        matcher = self.matcher(query)

        screen = self.screen
        assert isinstance(screen, MainScreen)

        for mode in sorted(
            screen.conversation.modes.values(), key=lambda mode: mode.name
        ):
            command = mode.name
            score = matcher.match(command)
            if score > 0:
                yield Hit(
                    score,
                    matcher.highlight(command),
                    partial(screen.conversation.set_mode, mode.id),
                    help=mode.description,
                )

    async def discover(self) -> Hits:
        screen = self.screen
        assert isinstance(screen, MainScreen)

        for mode in sorted(
            screen.conversation.modes.values(), key=lambda mode: mode.name
        ):
            yield DiscoveryHit(
                mode.name,
                partial(screen.conversation.set_mode, mode.id),
                help=mode.description,
            )


class MainScreen(Screen, can_focus=False):
    AUTO_FOCUS = "Conversation Prompt TextArea"

    CSS_PATH = "main.tcss"

    COMMANDS = {ModeProvider}

    SESSION_NAVIGATION_GROUP = Binding.Group(description="Sessions")
    BINDINGS = [
        Binding("ctrl+b,f20", "show_sidebar", "Sidebar"),
        Binding("ctrl+h", "go_home", "Home"),
        Binding(
            "ctrl+left_square_bracket",
            "session_previous",
            "Previous session",
            group=SESSION_NAVIGATION_GROUP,
        ),
        Binding(
            "ctrl+right_square_bracket",
            "session_next",
            "Next session",
            group=SESSION_NAVIGATION_GROUP,
        ),
    ]

    BINDING_GROUP_TITLE = "Screen"
    busy_count = var(0)
    throbber: getters.query_one[Throbber] = getters.query_one("#throbber")
    conversation = getters.query_one(Conversation)
    side_bar = getters.query_one(SideBar)
    project_directory_tree = getters.query_one("#project_directory_tree")

    column = reactive(False)
    column_width = reactive(100)
    scrollbar = reactive("")
    project_path: var[Path] = var(Path("./").expanduser().absolute())

    app = getters.app(ToadApp)

    def __init__(
        self,
        project_path: Path,
        agent: Agent | None = None,
        agent_session_id: str | None = None,
        agent_session_title: str | None = None,
        session_pk: int | None = None,
        initial_prompt: str | None = None,
    ) -> None:
        super().__init__()
        self.set_reactive(MainScreen.project_path, project_path)
        self._agent = agent
        self._agent_session_id = agent_session_id
        self._agent_session_title = agent_session_title
        self._session_pk = session_pk
        self._initial_prompt = initial_prompt

    def watch_title(self, title: str) -> None:
        self.app.update_terminal_title()

    def get_loading_widget(self) -> Widget:
        throbber = self.app.settings.get("ui.throbber", str)
        if throbber == "quotes":
            from toad.app import QUOTES
            from toad.widgets.future_text import FutureText

            quotes = QUOTES.copy()
            random.shuffle(quotes)
            return FutureText([Content(quote) for quote in quotes])
        return super().get_loading_widget()

    def _on_screen_resume(self, event: ScreenResume) -> None:
        self.conversation

    def compose(self) -> ComposeResult:
        with containers.Center():
            yield SideBar(
                SideBar.Panel("Plan", Plan([])),
                SideBar.Panel(
                    "Project",
                    ProjectDirectoryTree(
                        self.project_path,
                        id="project_directory_tree",
                    ),
                    flex=True,
                ),
            )
            yield Conversation(
                self.project_path,
                self._agent,
                self._agent_session_id,
                self._session_pk,
                initial_prompt=self._initial_prompt,
            ).data_bind(
                project_path=MainScreen.project_path,
                column=MainScreen.column,
            )
        yield Footer()

    def run_prompt(self, prompt: str) -> None:
        self.conversation

    def update_node_styles(self, animate: bool = True) -> None:
        self.conversation.update_node_styles(animate=animate)
        self.query_one(Footer).update_node_styles(animate=animate)
        self.query_one(SideBar).update_node_styles(animate=animate)

    def action_session_previous(self) -> None:
        if self.screen.id is not None:
            self.post_message(messages.SessionNavigate(self.screen.id, -1))

    def action_session_next(self) -> None:
        if self.screen.id is not None:
            self.post_message(messages.SessionNavigate(self.screen.id, +1))

    @on(messages.ProjectDirectoryUpdated)
    async def on_project_directory_update(self) -> None:
        await self.query_one(ProjectDirectoryTree).reload()

    @on(DirectoryTree.FileSelected, "ProjectDirectoryTree")
    def on_project_directory_tree_selected(self, event: Tree.NodeSelected):
        if (data := event.node.data) is not None:
            self.conversation.insert_path_into_prompt(data.path)

    @on(acp_messages.Plan)
    async def on_acp_plan(self, message: acp_messages.Plan):
        message.stop()
        entries = [
            Plan.Entry(
                Content(entry["content"]),
                entry.get("priority", "medium"),
                entry.get("status", "pending"),
            )
            for entry in message.entries
        ]
        self.query_one("SideBar Plan", Plan).entries = entries

    @on(messages.SessionUpdate)
    async def on_session_update(self, event: messages.SessionUpdate) -> None:
        # TODO: May not be required
        if event.name is not None:
            self._agent_session_title = event.name
        if self.id is not None:
            self.app.session_tracker.update_session(
                self.id,
                title=event.name,
                subtitle=event.subtitle,
                path=event.path,
                state=event.state,
            )

    @on(messages.SessionClose)
    async def on_session_close(self, event: messages.SessionClose) -> None:

        if self.id is None:
            return
        current_mode = self.id
        session_tracker = self.app.session_tracker

        session_count = session_tracker.session_count

        if session_count <= 1:

            session_tracker.close_session(current_mode)
            await self.app.switch_mode("store")

        else:
            if new_mode := self.app.session_tracker.session_cursor_move(
                current_mode, -1
            ):
                await self.app.switch_mode(new_mode)
            session_tracker.close_session(current_mode)

        self.app.call_later(self.app.remove_mode, current_mode)

    def on_mount(self) -> None:
        import gc

        gc.freeze()
        for tree in self.query("#project_directory_tree").results(DirectoryTree):
            tree.data_bind(path=MainScreen.project_path)
        for tree in self.query(DirectoryTree):
            tree.guide_depth = 3

    @on(OptionList.OptionHighlighted)
    def on_option_list_option_highlighted(
        self, event: OptionList.OptionHighlighted
    ) -> None:
        if event.option.id is not None:
            self.conversation.prompt.suggest(event.option.id)

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        if action == "show_sidebar" and self.side_bar.has_focus_within:
            return False
        return True

    def action_show_sidebar(self) -> None:
        self.side_bar.query_one("Collapsible CollapsibleTitle").focus()

    def action_focus_prompt(self) -> None:
        self.conversation.focus_prompt()

    async def action_go_home(self) -> None:
        await self.app.switch_mode("store")

    @on(SideBar.Dismiss)
    def on_side_bar_dismiss(self, message: SideBar.Dismiss):
        message.stop()
        self.conversation.focus_prompt()

    def watch_column(self, column: bool) -> None:
        self.conversation.styles.max_width = (
            max(10, self.column_width) if column else None
        )

    def watch_column_width(self, column_width: int) -> None:
        self.conversation.styles.max_width = (
            max(10, column_width) if self.column else None
        )

    def watch_scrollbar(self, old_scrollbar: str, scrollbar: str) -> None:
        if old_scrollbar:
            self.conversation.remove_class(f"-scrollbar-{old_scrollbar}")
        if scrollbar:
            self.conversation.add_class(f"-scrollbar-{scrollbar}")



import asyncio
import itertools
import threading
from pathlib import Path

from textual.cache import LRUCache
from textual.suggester import Suggester


class ListDirCache:
    """A cache for listing a directory (not a Rich / Textual object).

    This class is responsible for listing directories, and caching the results.

    Listing a directory is a blocking operation, which is why we defer the work to a thread.

    """

    def __init__(self) -> None:
        self._cache: LRUCache[tuple[str, int], list[Path]] = LRUCache(100)
        self._lock = threading.Lock()

    async def listdir(self, path: Path, size: int) -> list[Path]:
        cache_key = (str(path), size)

        def iterdir_thread(path: Path) -> list[Path]:
            """Run iterdir in a thread.

            Returns:
                A list of paths.
            """
            return list(itertools.islice(path.iterdir(), size))

        with self._lock:
            if cache_key in self._cache:
                paths = self._cache[cache_key]
            else:
                paths = await asyncio.to_thread(iterdir_thread, path)
                self._cache[cache_key] = paths
            return paths


class DirectorySuggester(Suggester):
    """Suggest a directory.

    This is a [Suggester](https://textual.textualize.io/api/suggester/#textual.suggester.Suggester) instance,
    used by the Input widget to suggest auto-completions.

    """

    def __init__(self) -> None:
        self._cache = ListDirCache()
        super().__init__(case_sensitive=True)

    async def get_suggestion(self, value: str) -> str | None:
        """Suggest the first matching directory."""

        try:
            path = Path(value)
            name = path.name

            children = await self._cache.listdir(
                path.expanduser() if path.is_dir() else path.parent.expanduser(), 100
            )
            possible_paths = [
                f"{sibling_path}/"
                for sibling_path in children
                if sibling_path.name.lower().startswith(name.lower())
                and sibling_path.is_dir()
            ]
            if possible_paths:
                possible_paths.sort(key=str.__len__)
                suggestion = possible_paths[0]

                if "~" in value:
                    home = str(Path("~").expanduser())
                    suggestion = suggestion.replace(home, "~", 1)
                return suggestion

        except FileNotFoundError:
            pass
        return None


if __name__ == "__main__":
    from textual.app import App, ComposeResult
    from textual import widgets

    class SApp(App):
        def compose(self) -> ComposeResult:
            yield widgets.Input(suggester=DirectorySuggester())

    SApp().run()



import asyncio
import os
from pathlib import Path
from typing import Literal


def longest_common_prefix(strings: list[str]) -> str:
    """
    Find the longest common prefix among a list of strings.

    Arguments:
        strings: List of strings

    Returns:
        The longest common prefix string
    """
    if not strings:
        return ""

    # Start with the first string as reference
    prefix: str = strings[0]

    # Compare with each subsequent string
    for current_string in strings[1:]:
        # Reduce prefix until it matches the start of current string
        while not current_string.startswith(prefix):
            prefix = prefix[:-1]
            if not prefix:
                return ""

    return prefix


class DirectoryReadTask:
    """A task to read a directory."""

    def __init__(self, path: Path) -> None:
        self.path = path
        self.done_event = asyncio.Event()
        self.directory_listing: list[Path] = []
        self._task: asyncio.Task | None = None

    def read(self) -> None:
        # TODO: Should this be cancellable, or have a maximum number of paths for the case of very large directories?
        for path in self.path.iterdir():
            self.directory_listing.append(path)

    def start(self) -> None:
        self._task = asyncio.create_task(
            self.run(), name=f"DirectoryReadTask({str(self.path)!r})"
        )

    async def run(self):
        try:
            await asyncio.to_thread(self.read)
        finally:
            self.done_event.set()

    async def wait(self) -> list[Path]:
        await self.done_event.wait()
        return self.directory_listing


class PathComplete:
    """Auto completes paths."""

    def __init__(self) -> None:
        self.read_tasks: dict[Path, DirectoryReadTask] = {}
        self.directory_listings: dict[Path, list[Path]] = {}

    async def __call__(
        self,
        current_working_directory: Path,
        path: str,
        *,
        exclude_type: Literal["file"] | Literal["dir"] | None = None,
    ) -> tuple[str | None, list[str] | None]:
        current_working_directory = (
            current_working_directory.expanduser().resolve().absolute()
        )
        directory_path = (current_working_directory / Path(path).expanduser()).resolve()

        node: str = path
        if not directory_path.is_dir():
            node = directory_path.name
            directory_path = directory_path.parent

        if (listing := self.directory_listings.get(directory_path)) is None:
            read_task = DirectoryReadTask(directory_path)
            self.read_tasks[directory_path] = read_task
            read_task.start()
            listing = await read_task.wait()

        if exclude_type is not None:
            if exclude_type == "dir":
                listing = [
                    listing_path
                    for listing_path in listing
                    if not listing_path.is_dir()
                ]
            else:
                listing = [
                    listing_path for listing_path in listing if listing_path.is_dir()
                ]

        if not node:
            return None, [listing_path.name for listing_path in listing]

        matching_nodes = [
            listing_path
            for listing_path in listing
            if listing_path.name.startswith(node)
        ]
        if not (matching_nodes):
            # Nothing matches
            return None, None

        if not (
            prefix := longest_common_prefix(
                [node_path.name for node_path in matching_nodes]
            )
        ):
            return None, None

        picked_path = directory_path / prefix
        path_size = (
            len(str(Path(directory_path).expanduser().resolve())) + 1 + len(node)
        )
        completed_prefix = str(picked_path)[path_size:]
        path_options = [
            str(path)[path_size + len(completed_prefix) :] for path in matching_nodes
        ]
        path_options = [name for name in path_options if name]

        if picked_path.is_dir() and not path_options:
            completed_prefix += os.sep

        return completed_prefix or None, path_options


if __name__ == "__main__":

    async def run():
        path_complete = PathComplete()
        cwd = Path("~/sandbox")

        print(await path_complete(cwd, "~/p"))

    asyncio.run(run())



from typing import Protocol, runtime_checkable, Iterable

from textual.widget import Widget

from toad.menus import MenuItem


@runtime_checkable
class BlockProtocol(Protocol):
    def block_cursor_up(self) -> Widget | None: ...
    def block_cursor_down(self) -> Widget | None: ...
    def get_cursor_block(self) -> Widget | None: ...
    def block_cursor_clear(self) -> None: ...
    def block_select(self, widget: Widget) -> None: ...


@runtime_checkable
class MenuProtocol(Protocol):
    def get_block_menu(self) -> Iterable[MenuItem]: ...
    def get_block_content(self, destination: str) -> str | None: ...


@runtime_checkable
class ExpandProtocol(Protocol):
    def can_expand(self) -> bool: ...
    def expand_block(self) -> None: ...
    def collapse_block(self) -> None: ...
    def is_block_expanded(self) -> bool: ...



from contextlib import suppress
from dataclasses import dataclass
from itertools import zip_longest
from pathlib import Path
from random import shuffle
from typing import Literal, Self

from textual.binding import Binding
from textual.screen import Screen
from textual import events
from textual import work
from textual import getters
from textual import on
from textual.app import ComposeResult
from textual.content import Content
from textual.css.query import NoMatches
from textual.message import Message
from textual.reactive import reactive
from textual import containers
from textual import widgets

import toad
from toad.app import ToadApp
from toad.format_path import format_path
from toad.pill import pill
from toad import messages
from toad.widgets.directory_input import DirectoryInput
from toad.widgets.mandelbrot import Mandelbrot
from toad.widgets.condensed_path import CondensedPath
from toad.widgets.grid_select import GridSelect
from toad.agent_schema import Agent
from toad.agents import read_agents


QR = """\
█▀▀▀▀▀█ ▄█ ▄▄█▄▄█ █▀▀▀▀▀█
█ ███ █ ▄█▀█▄▄█▄  █ ███ █
█ ▀▀▀ █ ▄ █ ▀▀▄▄▀ █ ▀▀▀ █
▀▀▀▀▀▀▀ ▀ ▀ ▀ █ █ ▀▀▀▀▀▀▀
█▀██▀ ▀█▀█▀▄▄█   ▀ █ ▀ █ 
 █ ▀▄▄▀▄▄█▄▄█▀██▄▄▄▄ ▀ ▀█
▄▀▄▀▀▄▀ █▀▄▄▄▀▄ ▄▀▀█▀▄▀█▀
█ ▄ ▀▀▀█▀ █ ▀ █▀ ▀ ██▀ ▀█
▀  ▀▀ ▀▀▄▀▄▄▀▀▄▀█▀▀▀█▄▀  
█▀▀▀▀▀█ ▀▄█▄▀▀  █ ▀ █▄▀▀█
█ ███ █ ██▄▄▀▀█▀▀██▀█▄██▄
█ ▀▀▀ █ ██▄▄ ▀  ▄▀ ▄▄█▀ █
▀▀▀▀▀▀▀ ▀▀▀  ▀   ▀▀▀▀▀▀▀▀"""


@dataclass
class ChangeDirectory(Message):
    path: str


class DirectoryDisplay(containers.HorizontalGroup):

    BINDINGS = [("escape", "dismiss", "Dismiss")]

    DEFAULT_CSS = """
    DirectoryDisplay {
        CondensedPath { display: block; }
        DirectoryInput { display: none; }
        &.-edit {
            CondensedPath { display: none}
            DirectoryInput { display: block; }
        }
    }
    """

    project_dir: reactive[Path] = reactive(Path)
    path = reactive("")
    edit = reactive(False, toggle_class="-edit")

    directory_input = getters.query_one(DirectoryInput)
    condensed_path = getters.query_one(CondensedPath)

    def __init__(self, project_dir: Path) -> None:
        super().__init__()
        self.project_dir = project_dir
        self.path = format_path(project_dir, directory=True)

    def watch_project_dir(self, path: Path) -> None:
        self.path = format_path(path, directory=True)

    def focus(self, scroll_visible=True) -> Self:
        self.edit = True
        self.directory_input.focus(scroll_visible=scroll_visible)
        return self

    @on(events.Click, "CondensedPath")
    def on_click(self) -> None:
        self.edit = True
        self.directory_input.focus()

    @on(events.DescendantBlur)
    def on_blur(self):
        self.action_dismiss()

    @on(widgets.Input.Submitted)
    def on_input_submitted(self, event: widgets.Input.Submitted) -> None:
        path = Path(event.value).expanduser().resolve()
        self.edit = False
        if not path.is_dir():
            self.notify(
                f"Unable to change directory to {str(path)!r}",
                title="Change directory",
                severity="error",
            )
            return
        self.condensed_path.path = format_path(path, directory=True)
        self.post_message(ChangeDirectory(str(path)))

    def action_dismiss(self) -> None:
        self.edit = False
        self.directory_input.value = self.path

    def watch_edit(self, edit: bool) -> None:
        if not edit and self.directory_input.has_focus:
            self.directory_input.blur()

    def compose(self) -> ComposeResult:
        yield widgets.Label("📁 ")
        yield CondensedPath(self.path, directory=True).data_bind(
            path=DirectoryDisplay.path
        ).with_tooltip("Project directory for new agent sessions (click to edit)")
        yield DirectoryInput(self.path, select_on_focus=True, compact=True).data_bind(
            value=DirectoryDisplay.path
        )


class AgentItem(containers.VerticalGroup):
    """An entry in the Agent grid select."""

    def __init__(self, agent: Agent) -> None:
        self._agent = agent
        super().__init__()

    @property
    def agent(self) -> Agent:
        return self._agent

    def compose(self) -> ComposeResult:
        agent = self._agent
        with containers.Grid():
            yield widgets.Label(agent["name"], id="name")
            tag = pill(agent["type"], "$primary-muted 50%", "$text-primary")
            yield widgets.Label(tag, id="type")
        yield widgets.Label(agent["author_name"], id="author")
        yield widgets.Static(agent["description"], id="description")


class LauncherGridSelect(GridSelect):

    HELP = """\
## Launcher

Your favorite agents.

- **1-9 a-f** Select agent
- **cursor keys** navigate agents
- **tab / shift+tab** Move to next / previous section
- **space** Launch highlighted agent
- **enter** Open agent details
"""
    BINDING_GROUP_TITLE = "Launcher"

    app = getters.app(ToadApp)
    BINDINGS = [
        Binding(
            "enter",
            "select",
            "Details",
            tooltip="Open agent details",
        ),
        Binding(
            "space",
            "launch",
            "Launch",
            tooltip="Launch highlighted agent",
        ),
    ]

    def action_details(self) -> None:
        if self.highlighted is None:
            return
        agent_item = self.children[self.highlighted]
        assert isinstance(agent_item, LauncherItem)
        self.post_message(StoreScreen.OpenAgentDetails(agent_item._agent["identity"]))

    def action_remove(self) -> None:
        agents = self.app.settings.get("launcher.agents", str).splitlines()
        if self.highlighted is None:
            return
        try:
            del agents[self.highlighted]
        except IndexError:
            pass
        else:
            self.app.settings.set("launcher.agents", "\n".join(agents))

    def action_launch(self) -> None:
        if self.highlighted is None:
            return
        child = self.children[self.highlighted]
        assert isinstance(child, LauncherItem)
        self.screen.post_message(messages.LaunchAgent(child.agent["identity"]))


class Launcher(containers.VerticalGroup):
    app = getters.app(ToadApp)
    grid_select = getters.query_one("#launcher-grid-select", LauncherGridSelect)
    DIGITS = "123456789ABCDEF"

    def __init__(
        self,
        agents: dict[str, Agent],
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self._agents = agents
        super().__init__(name=name, id=id, classes=classes)

    @property
    def highlighted(self) -> int | None:
        return self.grid_select.highlighted

    @highlighted.setter
    def highlighted(self, value: int) -> None:
        self.grid_select.highlighted = value

    def focus(self, scroll_visible: bool = True) -> Self:
        try:
            self.grid_select.focus(scroll_visible=scroll_visible)
        except NoMatches:
            pass
        return self

    def compose(self) -> ComposeResult:
        launcher_agents = list(
            dict.fromkeys(
                identity
                for identity in self.app.settings.get(
                    "launcher.agents", str
                ).splitlines()
                if identity.strip()
            )
        )
        agents = self._agents
        self.set_class(not launcher_agents, "-empty")
        if launcher_agents:
            with LauncherGridSelect(
                id="launcher-grid-select", min_column_width=32, max_column_width=32
            ):
                for digit, identity in zip_longest(self.DIGITS, launcher_agents):
                    if identity is None:
                        break
                    if identity in agents:
                        yield LauncherItem(digit or "", agents[identity])

        if not launcher_agents:
            yield widgets.Label("Choose your fighter below!", classes="no-agents")

    def launch_highlighted(self) -> None:
        self.grid_select.action_launch()


class LauncherItem(containers.VerticalGroup):
    """An entry in the Agent grid select."""

    def __init__(self, digit: str, agent: Agent) -> None:
        self._digit = digit
        self._agent = agent
        super().__init__()

    @property
    def agent(self) -> Agent:
        return self._agent

    def compose(self) -> ComposeResult:
        agent = self._agent
        with containers.HorizontalGroup():
            if self._digit:
                yield widgets.Digits(self._digit)
            with containers.VerticalGroup():
                yield widgets.Label(agent["name"], id="name")
                yield widgets.Label(agent["author_name"], id="author")
                yield widgets.Static(agent["description"], id="description")


class AgentGridSelect(GridSelect):
    HELP = """\
## Agent select

- **cursor keys** Navigate agents
- **tab / shift+tab** Move to next / previous section
- **enter** Open agent details
- **space** Launch the agent (if installed)
"""
    BINDINGS = [
        Binding("enter", "select", "Details", tooltip="Open agent details"),
        Binding("space", "launch", "Launch", tooltip="Launch highlighted agent"),
    ]
    BINDING_GROUP_TITLE = "Agent Select"

    def action_launch(self) -> None:
        if self.highlighted is None:
            return
        child = self.children[self.highlighted]
        if not isinstance(child, AgentItem):
            self.app.open_url("https://github.com/sponsors/willmcgugan")
            return
        self.post_message(messages.LaunchAgent(child.agent["identity"]))


class Container(containers.VerticalScroll):
    BINDING_GROUP_TITLE = "View"

    def allow_focus(self) -> bool:
        """Only allow focus when we can scroll."""
        return super().allow_focus() and self.show_vertical_scrollbar


class StoreScreen(Screen):
    BINDING_GROUP_TITLE = "Screen"
    CSS_PATH = "store.tcss"
    FOCUS_GROUP = Binding.Group("Focus")
    BINDINGS = [
        Binding(
            "tab",
            "app.focus_next",
            "Focus Next",
            group=FOCUS_GROUP,
        ),
        Binding(
            "shift+tab",
            "app.focus_previous",
            "Focus Previous",
            group=FOCUS_GROUP,
        ),
        Binding(
            "null",
            "quick_launch",
            "Quick launch",
            key_display="1-9 a-f",
        ),
        Binding("ctrl+r", "resume", "Resume", tooltip="Resume a previous session"),
        Binding(
            "ctrl+d",
            "directory",
            "Directory",
            tooltip="Change project directory",
        ),
    ]

    agents_view = getters.query_one("#agents-view", AgentGridSelect)
    launcher = getters.query_one("#launcher", Launcher)
    container = getters.query_one("#container", Container)

    project_dir: reactive[Path] = reactive(Path)

    app = getters.app(ToadApp)

    @dataclass
    class OpenAgentDetails(Message):
        identity: str

    def __init__(
        self, name: str | None = None, id: str | None = None, classes: str | None = None
    ):
        self._agents: dict[str, Agent] = {}
        super().__init__(name=name, id=id, classes=classes)
        self.project_dir = self.app.project_dir

    @property
    def agents(self) -> dict[str, Agent]:
        return self._agents

    def compose(self) -> ComposeResult:
        with containers.VerticalGroup(id="title-container"):
            with containers.Grid(id="title-grid"):
                yield Mandelbrot()
                yield widgets.Label(self.get_info(), id="info")
        yield DirectoryDisplay(self.project_dir).data_bind(
            project_dir=StoreScreen.project_dir
        )
        yield Container(id="container", can_focus=False)
        yield widgets.Footer()

    def get_info(self) -> Content:
        toad_version = toad.get_version()
        content = Content.assemble(
            Content.from_markup("🐸 Toad"),
            pill(f"v{toad_version}", "$primary-muted", "$text-primary"),
            ("\nThe universal interface for AI in your terminal", "$text-success"),
            (
                "\nSoftware lovingly crafted by hand (with a dash of AI) in Edinburgh, Scotland",
                "dim",
            ),
            "\n",
            (
                Content.from_markup(
                    "\nClick and hold to zoom the fractal, [b dim]ctrl+click[/] to zoom out."
                )
            ),
            "\n\n",
            (
                Content.from_markup(
                    "[dim]Code: [@click=screen.url('https://github.com/batrachianai/toad')]Repository[/] • "
                    "Bugs: [@click=screen.url('https://github.com/batrachianai/toad/discussions')]Discussions[/] • "
                    "Sponsor: [@click=screen.url('https://github.com/sponsors/willmcgugan')]@willmcgugan[/]"
                )
            ),
        )

        return content

    def action_url(self, url: str) -> None:
        import webbrowser

        webbrowser.open(url)

    def compose_agents(self) -> ComposeResult:
        agents = self._agents

        yield Launcher(agents, id="launcher")

        ordered_agents = sorted(
            agents.values(), key=lambda agent: agent["name"].casefold()
        )

        recommended_agents = [
            agent for agent in ordered_agents if agent.get("recommended", False)
        ]
        # Shuffle reccomended agents so none has priority
        shuffle(recommended_agents)
        if recommended_agents:
            with containers.VerticalGroup(id="sponsored-agents", classes="recommended"):
                yield widgets.Static(
                    "[$text-warning u]Recommended[/] [$text-secondary 100% i]Best of the bunch",
                    classes="heading",
                )
                with AgentGridSelect(classes="agents-picker", min_column_width=40):
                    for agent in recommended_agents:
                        yield AgentItem(agent)
                    yield widgets.Static(
                        "[$text-warning]Your agent here[/] — support development of Toad by [@click=screen.url('https://github.com/sponsors/willmcgugan')]sponsoring[/] this project",
                        classes="sponsor-me",
                    )

        chat_bots = [
            agent for agent in ordered_agents if agent["type"] in {"chat", "assistant"}
        ]
        if chat_bots:
            yield widgets.Static(
                "[$text-warning u]Chat & Assistants[/] [$text-secondary 100% i]Biddi-biddi-biddi",
                classes="heading",
            )
            with containers.VerticalGroup():
                with AgentGridSelect(classes="agents-picker", min_column_width=40):
                    for agent in chat_bots:
                        yield AgentItem(agent)

        coding_agents = [agent for agent in ordered_agents if agent["type"] == "coding"]
        if coding_agents:
            yield widgets.Static(
                "[$text-warning u]Coding agents[/] [$text-secondary i]Build software with AI",
                classes="heading",
            )
            with containers.VerticalGroup():
                with AgentGridSelect(classes="agents-picker", min_column_width=40):
                    for agent in coding_agents:
                        yield AgentItem(agent)

    def move_focus(self, direction: Literal[-1] | Literal[+1]) -> None:
        if isinstance(self.focused, GridSelect):
            focus_chain = list(self.query(GridSelect))
            if self.focused in focus_chain:
                index = focus_chain.index(self.focused)
                new_focus = focus_chain[(index + direction) % len(focus_chain)]
                if direction == -1:
                    new_focus.highlight_last()
                else:
                    new_focus.highlight_first()
                new_focus.focus(scroll_visible=False)

    @on(GridSelect.LeaveUp)
    def on_grid_select_leave_up(self, event: GridSelect.LeaveUp):
        event.stop()
        self.move_focus(-1)

    @on(GridSelect.LeaveDown)
    def on_grid_select_leave_down(self, event: GridSelect.LeaveUp):
        event.stop()
        self.move_focus(+1)

    @on(GridSelect.Selected, ".agents-picker")
    @work
    async def on_grid_select_selected(self, event: GridSelect.Selected):
        if not isinstance(event.widget, AgentItem):
            self.app.open_url("https://github.com/sponsors/willmcgugan")
            return
        assert isinstance(event.widget, AgentItem)
        from toad.screens.agent_modal import AgentModal

        modal_response = await self.app.push_screen_wait(AgentModal(event.widget.agent))
        await self.app.save_settings()
        if modal_response == "launch":
            self.post_message(messages.LaunchAgent(event.widget.agent["identity"]))

    @on(OpenAgentDetails)
    @work
    async def open_agent_detail(self, message: OpenAgentDetails) -> None:
        from toad.screens.agent_modal import AgentModal

        try:
            agent = self._agents[message.identity]
        except KeyError:
            return
        modal_response = await self.app.push_screen_wait(AgentModal(agent))
        await self.app.save_settings()
        if modal_response == "launch":
            self.post_message(messages.LaunchAgent(agent["identity"]))

    @on(GridSelect.Selected, "#launcher GridSelect")
    @work
    async def on_launcher_selected(self, event: GridSelect.Selected):
        launcher_item = event.widget
        assert isinstance(launcher_item, LauncherItem)

        from toad.screens.agent_modal import AgentModal

        modal_response = await self.app.push_screen_wait(
            AgentModal(launcher_item.agent)
        )
        await self.app.save_settings()
        if modal_response == "launch":
            self.post_message(messages.LaunchAgent(launcher_item.agent["identity"]))

    @on(ChangeDirectory)
    def on_change_directory(self, event: ChangeDirectory) -> None:
        self.project_dir = Path(event.path)
        self.app.project_dir = self.project_dir

    @work
    async def on_mount(self) -> None:
        self.app.settings_changed_signal.subscribe(self, self.setting_updated)
        try:
            self._agents = await read_agents()
        except Exception as error:
            self.notify(
                f"Failed to read agents data ({error})",
                title="Agents data",
                severity="error",
            )
        else:
            await self.container.mount_compose(self.compose_agents())
            with suppress(NoMatches):
                first_grid = self.container.query(GridSelect).first()
                first_grid.focus(scroll_visible=False)

    async def setting_updated(self, setting: tuple[str, object]) -> None:
        key, value = setting
        if key == "launcher.agents":
            await self.launcher.recompose()

            def focus_screen():
                try:
                    self.screen.query(GridSelect).focus()
                except Exception:
                    pass

            self.call_later(focus_screen)

    def on_key(self, event: events.Key) -> None:
        if event.character is None:
            return
        LAUNCHER_KEYS = "123456789abcdef"

        if event.character in LAUNCHER_KEYS:
            launch_item_offset = LAUNCHER_KEYS.find(event.character)
            try:
                self.launcher.grid_select.children[launch_item_offset]
            except IndexError:
                self.notify(
                    f"No agent on key [b]{LAUNCHER_KEYS[launch_item_offset]}",
                    title="Quick launch",
                    severity="error",
                )
                self.app.bell()
                return
            self.launcher.focus()
            self.launcher.highlighted = launch_item_offset
            self.launcher.launch_highlighted()

    def action_quick_launch(self) -> None:
        self.launcher.focus()

    @work
    async def action_resume(self) -> None:
        from toad.screens.session_resume_modal import SessionResumeModal

        session = await self.app.push_screen_wait(SessionResumeModal())
        if session is not None:
            self.post_message(
                messages.LaunchAgent(
                    session["agent_identity"],
                    session["agent_session_id"],
                    pk=session["id"],
                )
            )

    async def action_directory(self) -> None:
        if (directory_display := self.query_one_optional(DirectoryDisplay)) is not None:
            directory_display.focus()


if __name__ == "__main__":
    from toad.app import ToadApp

    app = ToadApp(mode="store")

    app.run()



from __future__ import annotations

from dataclasses import dataclass

from textual.app import ComposeResult
from textual.content import Content
from textual.reactive import reactive
from textual import containers
from textual.widgets import Static

from toad.pill import pill
from toad.widgets.strike_text import StrikeText


class NonSelectableStatic(Static):
    ALLOW_SELECT = False


class Plan(containers.Grid):
    # BORDER_TITLE = "Plan"
    DEFAULT_CLASSES = "block"
    DEFAULT_CSS = """

    Plan {        
        # border: panel $secondary;
        border-top: ascii $secondary;            
        border-bottom: ascii $secondary;
        background: black 10%;        
        margin: 1 0 1 0 !important;   # Special case because the tall border reduces the apparent width           
        padding: 0 1 0 1;        
        height: auto;                        
        grid-size: 2;
        grid-columns: auto 1fr;
        grid-rows: auto;
        border-title-align: center;

        .-no-plan {
            text-style: dim italic;
        }

        .plan {
            color: $text-secondary;
        }
        .status {
            padding: 0 0 0 0;
            color: $text-secondary;
        }
        .priority {
            padding: 0 0 0 0;
        }
        .status.status-completed {
            color: $text-success;            
        }
        .status-pending {
            opacity: 0.7;
        }          
    }

    """

    @dataclass(frozen=True)
    class Entry:
        """Information about an entry in the Plan."""

        content: Content
        priority: str
        status: str

        def update_status(self, status: str) -> Plan.Entry:
            """Get a new Entry with updated status.

            Args:
                status: New status

            Returns:
                New Entry instance.
            """
            return Plan.Entry(self.content, self.priority, status)

    entries: reactive[list[Entry] | None] = reactive(None, recompose=True)
    all_complete: reactive[bool] = reactive(False, toggle_class="-all-complete")

    PRIORITIES = {
        "high": pill("H", "$error-muted", "$text-error"),
        "medium": pill("M", "$warning-muted", "$text-warning"),
        "low": pill("L", "$primary-muted", "$text-primary"),
    }

    def __init__(
        self,
        entries: list[Entry],
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        self.newly_completed: set[Plan.Entry] = set()
        super().__init__(name=name, id=id, classes=classes)
        self.set_reactive(Plan.entries, entries)

    def watch_entries(self, old_entries: list[Entry], new_entries: list[Entry]) -> None:
        entry_map = {entry.content: entry for entry in old_entries}
        newly_completed: set[Plan.Entry] = set()
        for entry in new_entries:
            old_entry = entry_map.get(entry.content, None)
            if (
                old_entry is not None
                and entry.status == "completed"
                and entry.status != old_entry.status
            ):
                newly_completed.add(entry)
        self.newly_completed = newly_completed
        if new_entries:
            self.all_complete = all(
                entry.status == "completed" for entry in self.entries
            )
        else:
            self.all_complete = False

    def compose(self) -> ComposeResult:
        if not self.entries:
            yield Static("No plan yet", classes="-no-plan")
            return
        for entry in self.entries:
            classes = f"priority-{entry.priority} status-{entry.status}"
            yield NonSelectableStatic(
                self.render_status(entry.status),
                classes=f"status {classes}",
            )

            yield (
                strike_text := StrikeText(
                    entry.content,
                    classes=f"plan {classes}",
                )
            )
            if entry in self.newly_completed:
                self.call_after_refresh(strike_text.strike)
            elif entry.status == "completed":
                strike_text.add_class("-complete")
        self.all_complete = all(entry.status == "completed" for entry in self.entries)

    def render_status(self, status: str) -> Content:
        if status == "completed":
            return Content(" ✔ ")
        elif status == "pending":
            return Content(" • ")
        elif status == "in_progress":
            return Content("👉 ")
        return Content()


if __name__ == "__main__":
    from textual.app import App

    entries = [
        Plan.Entry(
            Content("Build the best damn UI for agentic coding in the terminal"),
            "hide",
            "in_progress",
        ),
        Plan.Entry(
            Content(
                "Embarass big tech by being the only agent CLI that can render Markdown tables"
            ),
            "high",
            "pending",
        ),
        Plan.Entry(
            Content.from_markup("Catch flight to Wuhan"),
            "low",
            "pending",
        ),
        Plan.Entry(
            Content.from_markup("Eat 热干面 for breakfast"),
            "low",
            "pending",
        ),
        Plan.Entry(
            Content.from_markup("Pack sunscreen, catch flight to Thailand"),
            "low",
            "pending",
        ),
        Plan.Entry(
            Content.from_markup(
                "Work as a digital nomad in Asia, eat well, don't get sun-stroke"
            ),
            "low",
            "pending",
        ),
    ]

    class PlanApp(App):
        BINDINGS = [("space", "strike")]

        CSS = """
        Screen {
            align: center middle;
            Plan {
                margin: 1;
            }
        }
        """

        def __init__(self) -> None:
            super().__init__()
            self.working = 0

        def compose(self) -> ComposeResult:
            yield Plan(entries)

        def action_strike(self) -> None:
            new_entries = entries.copy()
            new_entries[self.working] = entries[self.working].update_status("completed")
            self.working += 1
            try:
                new_entries[self.working] = entries[self.working].update_status(
                    "in_progress"
                )
            except IndexError:
                pass

            self.query_one(Plan).entries = new_entries
            entries[:] = new_entries

    app = PlanApp()
    app.run()



# /// script
# dependencies = [
#   "qrcode",
# ]
# ///

import qrcode

qr = qrcode.QRCode(
    border=0,
    version=1,
    error_correction=qrcode.constants.ERROR_CORRECT_L,
)
# qr.add_data('https://github.com/sponsors/willmcgugan')
qr.add_data("https://tinyurl.com/52r7fd25")
qr.make(fit=True)

# Print using Unicode blocks (looks more square)
qr.print_ascii()



from functools import lru_cache
from time import monotonic

from rich.segment import Segment
from rich.style import Style as RichStyle

from textual.visual import Visual
from textual.color import Color, Gradient

from textual.style import Style
from textual.strip import Strip
from textual.visual import RenderOptions
from textual.widget import Widget
from textual.css.styles import RulesMap


COLORS = [
    "#881177",
    "#aa3355",
    "#cc6666",
    "#ee9944",
    "#eedd00",
    "#99dd55",
    "#44dd88",
    "#22ccbb",
    "#00bbcc",
    "#0099cc",
    "#3366bb",
    "#663399",
    "#881177",
]


class ThrobberVisual(Visual):
    """A Textual 'Visual' object.

    Analogous to a Rich renderable, but with support for transparency.

    """

    def __init__(
        self, character: str = "━", get_time: Callable[[], float] = monotonic
    ) -> None:
        self.character = character
        self.get_time = get_time

    gradient = Gradient.from_colors(*[Color.parse(color) for color in COLORS])

    @lru_cache(maxsize=8)
    def make_segments(self, style: Style, width: int) -> list[Segment]:
        gradient = self.gradient
        background = style.rich_style.bgcolor
        character = self.character
        segments = [
            Segment(
                character,
                RichStyle.from_color(
                    gradient.get_rich_color((offset / width) % 1),
                    background,
                ),
            )
            for offset in range(width * 2)
        ]

        return segments

    def render_strips(
        self, width: int, height: int | None, style: Style, options: RenderOptions
    ) -> list[Strip]:
        """Render the Visual into an iterable of strips.

        Args:
            width: Width of desired render.
            height: Height of desired render or `None` for any height.
            style: The base style to render on top of.
            options: Additional render options.

        Returns:
            An list of Strips.
        """

        time = self.get_time()
        segments = self.make_segments(style, width)
        offset = width - int((time % 1.0) * width)
        segments = segments[offset : offset + width]
        return [Strip(segments, cell_length=width)]

    def get_optimal_width(self, rules: RulesMap, container_width: int) -> int:
        return container_width

    def get_height(self, rules: RulesMap, width: int) -> int:
        return 1


class Throbber(Widget):
    def on_mount(self) -> None:
        self.auto_refresh = 1 / 15

    def render(self) -> ThrobberVisual:
        return ThrobberVisual()



from enum import IntEnum
from functools import lru_cache
from pathlib import Path
from typing import Iterable, NamedTuple, Sequence

from textual.content import Span


SAFE_COMMANDS = {
    # Display & Output
    "echo",
    "cat",
    "less",
    "more",
    "head",
    "tail",
    "tac",
    "nl",
    # File & Directory Information
    "ls",
    "tree",
    "pwd",
    "file",
    "stat",
    "du",
    "df",
    # Search & Find
    "find",
    "locate",
    "which",
    "whereis",
    "type",
    "grep",
    "egrep",
    "fgrep",
    # Text Processing (read-only)
    "wc",
    "sort",
    "uniq",
    "cut",
    "paste",
    "column",
    "tr",
    "diff",
    "cmp",
    "comm",
    # System Information
    "whoami",
    "who",
    "w",
    "id",
    "hostname",
    "uname",
    "uptime",
    "date",
    "cal",
    "env",
    "printenv",
    # Process Information
    "ps",
    "top",
    "htop",
    "pgrep",
    "jobs",
    "pstree",
    # Network (read-only operations)
    "ping",
    "traceroute",
    "nslookup",
    "dig",
    "host",
    "netstat",
    "ss",
    "ifconfig",
    "ip",
    # View compressed files (without extracting)
    "zcat",
    "zless",
    # History & Help
    "history",
    "man",
    "help",
    "info",
    "apropos",
    "whatis",
    # Comparison & Checksums
    "md5sum",
    "sha256sum",
    "sha1sum",
    "cksum",
    "sum",
    # Other Safe Commands
    "bc",
    "expr",
    "test",
    "sleep",
    "true",
    "false",
    "yes",
    "seq",
    "basename",
    "dirname",
    "realpath",
    "readlink",
}

UNSAFE_COMMANDS = {
    # File/Directory Creation
    "mkdir",
    "touch",
    "mktemp",
    "mkfifo",
    "mknod",
    # File/Directory Deletion
    "rm",
    "rmdir",
    "shred",
    # File/Directory Moving/Copying
    "mv",
    "cp",
    "rsync",
    "scp",
    "install",
    # File Modification/Editing
    "sed",  # with -i flag
    "awk",  # can write files
    "tee",  # writes to files and stdout
    # Permissions/Ownership
    "chmod",
    "chown",
    "chgrp",
    "chattr",
    "setfacl",
    # Linking
    "ln",
    "link",
    "unlink",
    # Archive/Compression (extract/compress operations)
    "tar",
    "untar",
    "zip",
    "unzip",
    "gzip",
    "gunzip",
    "bzip2",
    "bunzip2",
    "xz",
    "unxz",
    "7z",
    "rar",
    "unrar",
    # Download Tools
    "wget",
    "curl",
    "fetch",
    "aria2c",
    # Low-level Disk Operations
    "dd",
    "truncate",
    "fallocate",
    # File Splitting
    "split",
    "csplit",
    # Synchronization
    "sync",
    # System Administration
    "useradd",
    "userdel",
    "usermod",
    "groupadd",
    "groupdel",
    "passwd",
    "mount",
    "umount",
    "mkfs",
    "fdisk",
    "parted",
    "swapon",
    "swapoff",
    # Other Potentially Dangerous
    "patch",
}


COMMAND_SPLIT = {";", "&&", "||", "|"}
CHANGE_DIRECTORY = {"cd"}


class DangerLevel(IntEnum):
    """The danger level of a command."""

    SAFE = 0  # Command is know to be generally save
    UNKNOWN = 1  # We don't know about this command
    DANGEROUS = 2  # Command is potentially dangerous (can modify filesystem)
    DESTRUCTIVE = 3  # Command is both dangerous and refers outside of project root


class CommandAtom(NamedTuple):
    """A command "atom"."""

    name: str
    """Name of the command."""
    level: DangerLevel
    """Danger level."""
    path: Path
    """The path to which this command is expected to apply."""
    span: tuple[int, int]
    """Span to highlight error."""


@lru_cache(maxsize=1024)
def detect(
    project_directory: str,
    current_working_directory: str,
    command_line: str,
    *,
    danger_style: str = "",
    destructive_style: str = "$text-error on $error-muted 70%",
) -> tuple[Sequence[Span], DangerLevel]:
    """Attempt to detect potentially destructive commands.

    Args:
        project_directory: Project directory.
        current_working_directory: Current working directory.
        command_line: Bash command.
        danger_style: Style to highlight dangerous commands.
        destructive_style: Style highlight destructive commands.

    Returns:
        A tuple of spans to highlight the command, and a `DangerLevel` enumeration.
    """
    try:
        atoms = list(
            analyze(project_directory, current_working_directory, command_line)
        )
    except OSError:
        return [], DangerLevel.UNKNOWN
    spans: list[Span] = []
    for atom in atoms:
        if atom.level == DangerLevel.DANGEROUS and danger_style:
            spans.append(Span(*atom.span, danger_style))
        elif atom.level == DangerLevel.DESTRUCTIVE and destructive_style:
            spans.append(Span(*atom.span, destructive_style))

    if atoms:
        danger_level = max(command_atom.level for command_atom in atoms)
    else:
        danger_level = DangerLevel.SAFE

    return (spans, danger_level)


def analyze(
    project_directory: str, current_working_directory: str, command_line: str
) -> Iterable[CommandAtom]:
    """Analyze a command and generate information about potentially destructive commands.

    Args:
        project_dir: The project directory.
        command_line: A bash command line.

    Yields:
        `CommandAtom` objects.
    """
    project_path = Path(project_directory).resolve()

    import bashlex
    from bashlex import ast

    def recurse_nodes(root_path: Path, nodes: list[ast.node]) -> Iterable[CommandAtom]:
        for node in nodes:
            kind: str = node.kind

            if kind == "list":
                yield from recurse_nodes(root_path, node.parts)
                return

            if kind == "operator":
                continue

            level = DangerLevel.UNKNOWN

            if not hasattr(node, "parts"):
                continue

            if node.parts:
                command_name = command_line[slice(*node.parts[0].pos)]
                if command_name in SAFE_COMMANDS:
                    level = DangerLevel.SAFE
                elif command_name in UNSAFE_COMMANDS:
                    level = DangerLevel.DANGEROUS
                parts = node.parts[1:]
            else:
                parts = node.parts
                command_name = ""

            if not parts:
                yield CommandAtom(command_name, level, root_path, node.pos)
                continue

            change_directory = command_name in CHANGE_DIRECTORY

            for command_node in parts:
                command_word = command_line[slice(*node.pos)]

                if command_node.kind == "redirect":
                    redirect = command_line[slice(*command_node.output.pos)]
                    try:
                        target_path = (
                            root_path / Path(redirect).expanduser()
                        ).resolve()
                    except OSError:
                        continue
                    if not target_path.is_relative_to(project_path):
                        yield CommandAtom(
                            "redirect",
                            DangerLevel.DESTRUCTIVE,
                            target_path,
                            command_node.pos,
                        )
                    continue

                if command_node.kind == "command":
                    yield from recurse_nodes(root_path, command_node.parts)
                    continue
                if command_word.startswith(("-", "+")):
                    continue
                word = command_line[slice(*command_node.pos)]
                if change_directory:
                    try:
                        root_path = (root_path / Path(word)).expanduser().resolve()
                    except OSError:
                        pass
                    continue

                try:
                    target_path = (root_path / Path(word)).expanduser().resolve()
                except OSError:
                    continue
                if level == DangerLevel.DANGEROUS and not target_path.is_relative_to(
                    project_path
                ):
                    # If refers to a path outside of the project, upgrade to destructive
                    level = DangerLevel.DESTRUCTIVE

                yield CommandAtom(command_word, level, target_path, node.pos)

    current_path = Path(current_working_directory)
    try:
        nodes = bashlex.parse(command_line)
    except Exception:
        # Failed to parse bash
        return

    yield from recurse_nodes(current_path, nodes)


if __name__ == "__main__":
    import os
    from rich import print

    TEST = [
        "ls;ls",
        "echo 'hello world'",
        "rm foo",
        "rm ../foo",
        "rm /",
        "cat foo > ../foo.txt",
    ]

    for test in TEST:
        print(repr(test), detect(os.getcwd(), os.getcwd(), test))



from __future__ import annotations

from dataclasses import dataclass

from asyncio import Future
from typing import Any, Mapping, TYPE_CHECKING
from textual.message import Message

import rich.repr

from toad.answer import Answer
from toad.acp import protocol
from toad.acp.encode_tool_call_id import encode_tool_call_id
from toad.acp.agent import Mode

if TYPE_CHECKING:
    from toad.widgets.terminal_tool import ToolState


class AgentMessage(Message):
    pass


@dataclass
class Thinking(AgentMessage):
    type: str
    text: str


@dataclass
class UpdateStatusLine(AgentMessage):
    status_line: str


@dataclass
class Update(AgentMessage):
    type: str
    text: str


@dataclass
class UserMessage(Message):
    type: str
    text: str


@dataclass
@rich.repr.auto
class RequestPermission(AgentMessage):
    options: list[protocol.PermissionOption]
    tool_call: protocol.ToolCallUpdatePermissionRequest
    result_future: Future[Answer]


@dataclass
class Plan(AgentMessage):
    entries: list[protocol.PlanEntry]


@dataclass
class ToolCall(AgentMessage):
    tool_call: protocol.ToolCall

    @property
    def tool_id(self) -> str:
        """An id suitable for use as a TCSS ID."""
        return encode_tool_call_id(self.tool_call["toolCallId"])


@dataclass
class ToolCallUpdate(AgentMessage):
    tool_call: protocol.ToolCall
    update: protocol.ToolCallUpdate

    @property
    def tool_id(self) -> str:
        """An id suitable for use as a TCSS ID."""
        return encode_tool_call_id(self.tool_call["toolCallId"])


@dataclass
class AvailableCommandsUpdate(AgentMessage):
    """The agent is reporting its slash commands."""

    commands: list[protocol.AvailableCommand]


@dataclass
class CreateTerminal(AgentMessage):
    """Request a terminal in the conversation."""

    terminal_id: str
    command: str
    result_future: Future[bool]
    args: list[str] | None = None
    cwd: str | None = None
    env: Mapping[str, str] | None = None
    output_byte_limit: int | None = None


@dataclass
class KillTerminal(AgentMessage):
    """Kill a terminal process."""

    terminal_id: str


@dataclass
class GetTerminalState(AgentMessage):
    """Get the state of the terminal."""

    terminal_id: str
    result_future: Future[ToolState]


@dataclass
class ReleaseTerminal(AgentMessage):
    """Release the terminal."""

    terminal_id: str


@dataclass
class WaitForTerminalExit(AgentMessage):
    """Wait for the terminal to exit."""

    terminal_id: str
    result_future: Future[tuple[int, str | None]]


@rich.repr.auto
@dataclass
class SetModes(AgentMessage):
    """Set modes from agent."""

    current_mode: str
    modes: dict[str, Mode]


@dataclass
class ModeUpdate(AgentMessage):
    """Agent informed us about a mode change."""

    current_mode: str



import tempfile
import os


class AtomicWriteError(Exception):
    """An Atomic write failed."""


def write(path: str, content: str) -> None:
    """Write a file in an atomic manner.

    Args:
        filename: Filename of new file.
        content: Content to write.

    """
    path = os.path.abspath(path)
    dir_name = os.path.dirname(path) or "."
    try:
        with tempfile.NamedTemporaryFile(
            mode="w",
            encoding="utf-8",
            delete=False,
            dir=dir_name,
            prefix=f".{os.path.basename(path)}_tmp_",
        ) as temporary_file:
            temporary_file.write(content)
            temp_name = temporary_file.name
    except Exception as error:
        raise AtomicWriteError(
            f"Failed to write {path!r}; error creating temporary file: {error}"
        )

    try:
        os.replace(temp_name, path)  # Atomic on POSIX and Windows
    except Exception as error:
        raise AtomicWriteError(f"Failed to write {path!r}; {error}")
